"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const _ = require("lodash");
const fs = require("fs");
const browsers_1 = require("@puppeteer/browsers");
const grpc_plugin_1 = require("./plugin/v2/grpc_plugin");
const http_server_1 = require("./service/http-server");
const config_1 = require("./service/config");
const config_2 = require("./plugin/v2/config");
const logger_1 = require("./logger");
const minimist = require("minimist");
const node_plugin_1 = require("./node-plugin");
const Sanitizer_1 = require("./sanitizer/Sanitizer");
async function main() {
    const argv = minimist(process.argv.slice(2));
    const env = Object.assign({}, process.env);
    const command = argv._[0];
    // See https://github.com/grafana/grafana-image-renderer/issues/460
    process.env["PUPPETEER_DISABLE_HEADLESS_WARNING"] = "true";
    if (command === undefined) {
        const logger = new logger_1.PluginLogger();
        const config = config_2.defaultPluginConfig;
        (0, config_2.populatePluginConfigFromEnv)(config, env);
        if (!config.rendering.chromeBin && process.pkg) {
            const execPath = path.dirname(process.execPath);
            const chromeInfoFile = fs.readFileSync(path.resolve(execPath, 'chrome-info.json'), 'utf8');
            const chromeInfo = JSON.parse(chromeInfoFile);
            config.rendering.chromeBin = (0, browsers_1.computeExecutablePath)({
                cacheDir: path.dirname(process.execPath),
                browser: browsers_1.Browser.CHROME,
                buildId: chromeInfo.buildId,
            });
            logger.debug(`Setting chromeBin to ${config.rendering.chromeBin}`);
        }
        await (0, node_plugin_1.serve)({
            handshakeConfig: {
                protocolVersion: 2,
                magicCookieKey: 'grafana_plugin_type',
                magicCookieValue: 'datasource',
            },
            versionedPlugins: {
                2: {
                    renderer: new grpc_plugin_1.RenderGRPCPluginV2(config, logger),
                },
            },
            logger: logger,
            grpcHost: config.plugin.grpc.host,
            grpcPort: config.plugin.grpc.port,
        });
    }
    else if (command === 'server') {
        let config = config_1.defaultServiceConfig;
        if (argv.config) {
            try {
                const fileConfig = readJSONFileSync(argv.config);
                config = _.merge(config, fileConfig);
            }
            catch (e) {
                console.error('failed to read config from path', argv.config, 'error', e);
                return;
            }
        }
        (0, config_1.populateServiceConfigFromEnv)(config, env);
        const logger = new logger_1.ConsoleLogger(config.service.logging);
        const sanitizer = (0, Sanitizer_1.createSanitizer)();
        const server = new http_server_1.HttpServer(config, logger, sanitizer);
        await server.start();
    }
    else {
        console.log('Unknown command');
    }
}
main().catch((err) => {
    const errorLog = {
        '@level': 'error',
        '@message': 'failed to start grafana-image-renderer',
        'error': err.message,
        'trace': err.stack,
    };
    console.error(JSON.stringify(errorLog));
    process.exit(1);
});
function readJSONFileSync(filePath) {
    const rawdata = fs.readFileSync(filePath, 'utf8');
    return JSON.parse(rawdata);
}
;
//# sourceMappingURL=app.js.map