"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.serve = exports.defaultGRPCServer = exports.healthProtoDescriptor = exports.healthPackageDef = void 0;
const grpc = require("@grpc/grpc-js");
const protoLoader = require("@grpc/proto-loader");
const types_1 = require("./types");
const logger_1 = require("../../logger");
exports.healthPackageDef = protoLoader.loadSync(__dirname + '/../../../proto/health.proto', {
    keepCase: true,
    longs: String,
    enums: String,
    defaults: true,
    oneofs: true,
});
exports.healthProtoDescriptor = grpc.loadPackageDefinition(exports.healthPackageDef);
const defaultGRPCServer = () => new grpc.Server();
exports.defaultGRPCServer = defaultGRPCServer;
const protocolVersion = (opts) => {
    let protoVersion = opts.handshakeConfig.protocolVersion;
    const pluginSet = opts.plugins;
    const env = Object.assign({}, process.env);
    let clientVersions = [];
    if (env['PLUGIN_PROTOCOL_VERSIONS']) {
        const protocolVersions = env['PLUGIN_PROTOCOL_VERSIONS'].split(',');
        for (let n = 0; n < protocolVersions.length; n++) {
            const protocolVersion = parseInt(protocolVersions[n], 10);
            clientVersions.push(protocolVersion);
        }
    }
    // We want to iterate in reverse order, to ensure we match the newest
    // compatible plugin version.
    clientVersions = clientVersions.sort().reverse();
    // set the old un-versioned fields as if they were versioned plugins
    if (!opts.versionedPlugins) {
        opts.versionedPlugins = {};
    }
    if (pluginSet) {
        opts.versionedPlugins[protoVersion] = pluginSet;
    }
    // Sort the versions to make sure we match the latest first
    let versions = [];
    for (let n = 0; n < Object.keys(opts.versionedPlugins).length; n++) {
        const version = Object.keys(opts.versionedPlugins)[n];
        versions.push(parseInt(version, 10));
    }
    versions = versions.sort().reverse();
    let versionedPluginSet = {};
    for (let n = 0; n < versions.length; n++) {
        const version = versions[n];
        // Record each version, since we guarantee that this returns valid
        // values even if they are not a protocol match.
        protoVersion = version;
        versionedPluginSet = opts.versionedPlugins[version];
        for (let i = 0; i < clientVersions.length; i++) {
            const clientVersion = clientVersions[i];
            if (clientVersion === protoVersion) {
                return {
                    protoVersion,
                    pluginSet: versionedPluginSet,
                };
            }
        }
    }
    return {
        protoVersion,
        pluginSet: versionedPluginSet,
    };
};
const serve = async (opts) => {
    const env = Object.assign({}, process.env);
    opts.logger = opts.logger || new logger_1.PluginLogger();
    if (opts.handshakeConfig.magicCookieKey === '' || opts.handshakeConfig.magicCookieValue === '') {
        throw new Error('Misconfigured ServeConfig given to serve this plugin: no magic cookie key or value was set. Please notify the plugin author and report this as a bug.');
    }
    if (env[opts.handshakeConfig.magicCookieKey] !== opts.handshakeConfig.magicCookieValue) {
        throw new Error('This binary is a plugin. These are not meant to be executed directly. Please execute the program that consumes these plugins, which will load any plugins automatically');
    }
    // negotiate the version and plugins
    // start with default version in the handshake config
    const { protoVersion, pluginSet } = protocolVersion(opts);
    const server = new grpc.Server();
    const grpcHealthV1 = exports.healthProtoDescriptor['grpc']['health']['v1'];
    server.addService(grpcHealthV1.Health.service, {
        check: (_, callback) => {
            callback(null, { status: 'SERVING' });
        },
    });
    // Register all plugins onto the gRPC server.
    for (const key in pluginSet) {
        if (pluginSet.hasOwnProperty(key)) {
            const p = pluginSet[key];
            await p.grpcServer(server);
        }
    }
    opts.grpcHost = opts.grpcHost || '127.0.0.1';
    opts.grpcPort = opts.grpcPort || 0;
    return new Promise((resolve, reject) => {
        const address = `${opts.grpcHost}:${opts.grpcPort}`;
        server.bindAsync(address, grpc.ServerCredentials.createInsecure(), (error, port) => {
            if (error) {
                reject(error);
            }
            if (port === 0) {
                reject(new Error(`failed to bind address=${address}, boundPortNumber=${port}`));
            }
            server.start();
            console.log(`${types_1.coreProtocolVersion}|${protoVersion}|tcp|${opts.grpcHost}:${port}|grpc`);
            resolve();
        });
    });
};
exports.serve = serve;
//# sourceMappingURL=server.js.map