"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.HttpServer = void 0;
const bodyParser = require("body-parser");
const boom = require("@hapi/boom");
const contentDisposition = require("content-disposition");
const express = require("express");
const fs = require("fs");
const http = require("http");
const https = require("https");
const morgan = require("morgan");
const multer = require("multer");
const path = require("path");
const promClient = require("prom-client");
const browser_1 = require("../browser");
const metrics_1 = require("./metrics");
const types_1 = require("../sanitizer/types");
const middlewares_1 = require("./middlewares");
const upload = multer({ storage: multer.memoryStorage() });
var SanitizeRequestPartName;
(function (SanitizeRequestPartName) {
    SanitizeRequestPartName["file"] = "file";
    SanitizeRequestPartName["config"] = "config";
})(SanitizeRequestPartName || (SanitizeRequestPartName = {}));
class HttpServer {
    constructor(config, log, sanitizer) {
        this.config = config;
        this.log = log;
        this.sanitizer = sanitizer;
        this.render = async (req, res, next) => {
            if (!req.query.url) {
                throw boom.badRequest('Missing url parameter');
            }
            const headers = {};
            if (req.headers['Accept-Language']) {
                headers['Accept-Language'] = req.headers['Accept-Language'].join(';');
            }
            const options = {
                url: req.query.url,
                width: req.query.width,
                height: req.query.height,
                filePath: req.query.filePath,
                timeout: req.query.timeout,
                renderKey: req.query.renderKey,
                domain: req.query.domain,
                timezone: req.query.timezone,
                encoding: req.query.encoding,
                deviceScaleFactor: req.query.deviceScaleFactor,
                headers: headers,
            };
            this.log.debug('Render request received', 'url', options.url);
            req.on('close', (err) => {
                this.log.debug('Connection closed', 'url', options.url, 'error', err);
            });
            const result = await this.browser.render(options);
            res.sendFile(result.filePath, (err) => {
                if (err) {
                    next(err);
                }
                else {
                    try {
                        this.log.debug('Deleting temporary file', 'file', result.filePath);
                        fs.unlinkSync(result.filePath);
                    }
                    catch (e) {
                        this.log.error('Failed to delete temporary file', 'file', result.filePath);
                    }
                }
            });
        };
        this.sanitize = async (req, res) => {
            const file = req.files?.[SanitizeRequestPartName.file]?.[0];
            if (!file) {
                throw boom.badRequest('missing file');
            }
            const configFile = req.files?.[SanitizeRequestPartName.config]?.[0];
            if (!configFile) {
                throw boom.badRequest('missing config');
            }
            const config = JSON.parse(configFile.buffer.toString());
            const sanitizeReq = {
                ...config,
                content: file.buffer,
            };
            if (!(0, types_1.isSanitizeRequest)(sanitizeReq)) {
                throw boom.badRequest('invalid request: ' + JSON.stringify(config));
            }
            this.log.debug('Sanitize request received', 'contentLength', file.size, 'name', file.filename, 'config', JSON.stringify(config));
            try {
                const sanitizeResponse = this.sanitizer.sanitize(sanitizeReq);
                res.writeHead(200, {
                    'Content-Disposition': `attachment;filename=${file.filename ?? 'sanitized'}`,
                    'Content-Length': sanitizeResponse.sanitized.length,
                    'Content-Type': file.mimetype ?? 'application/octet-stream',
                });
                return res.end(sanitizeResponse.sanitized);
            }
            catch (e) {
                this.log.error('Sanitization failed', 'filesize', file.size, 'name', file.filename, 'error', e.stack);
                return res.status(500).json({ error: e.message });
            }
        };
        this.renderCSV = async (req, res, next) => {
            if (!req.query.url) {
                throw boom.badRequest('Missing url parameter');
            }
            const headers = {};
            if (req.headers['Accept-Language']) {
                headers['Accept-Language'] = req.headers['Accept-Language'].join(';');
            }
            const options = {
                url: req.query.url,
                filePath: req.query.filePath,
                timeout: req.query.timeout,
                renderKey: req.query.renderKey,
                domain: req.query.domain,
                timezone: req.query.timezone,
                encoding: req.query.encoding,
                headers: headers,
            };
            this.log.debug('Render request received', 'url', options.url);
            req.on('close', (err) => {
                this.log.debug('Connection closed', 'url', options.url, 'error', err);
            });
            const result = await this.browser.renderCSV(options);
            if (result.fileName) {
                res.setHeader('Content-Disposition', contentDisposition(result.fileName));
            }
            res.sendFile(result.filePath, (err) => {
                if (err) {
                    next(err);
                }
                else {
                    try {
                        this.log.debug('Deleting temporary file', 'file', result.filePath);
                        fs.unlink(result.filePath, (err) => {
                            if (err) {
                                throw err;
                            }
                            if (!options.filePath) {
                                fs.rmdir(path.dirname(result.filePath), () => { });
                            }
                        });
                    }
                    catch (e) {
                        this.log.error('Failed to delete temporary file', 'file', result.filePath, 'error', e.message);
                    }
                }
            });
        };
    }
    async start() {
        this.app = express();
        this.app.use(morgan('combined', {
            skip: (req, res) => {
                return res.statusCode >= 400;
            },
            stream: this.log.debugWriter,
        }));
        this.app.use(morgan('combined', {
            skip: (req, res) => {
                return res.statusCode < 400;
            },
            stream: this.log.errorWriter,
        }));
        this.app.use(bodyParser.json());
        if (this.config.service.metrics.enabled) {
            (0, metrics_1.setupHttpServerMetrics)(this.app, this.config.service.metrics, this.log);
        }
        this.app.get('/', (req, res) => {
            res.send('Grafana Image Renderer');
        });
        // Middlewares for /render endpoints
        this.app.use('/render', (0, middlewares_1.authTokenMiddleware)(this.config.service.security), middlewares_1.trustedUrlMiddleware);
        // Set up /render endpoints
        this.app.get('/render', (0, middlewares_1.asyncMiddleware)(this.render));
        this.app.get('/render/csv', (0, middlewares_1.asyncMiddleware)(this.renderCSV));
        this.app.get('/render/version', (req, res) => {
            const pluginInfo = require('../../plugin.json');
            res.send({ version: pluginInfo.info.version });
        });
        // Middlewares for /sanitize endpoints
        this.app.use('/sanitize', (0, middlewares_1.authTokenMiddleware)(this.config.service.security));
        // Set up /sanitize endpoints
        this.app.post('/sanitize', upload.fields([
            { name: SanitizeRequestPartName.file, maxCount: 1 },
            { name: SanitizeRequestPartName.config, maxCount: 1 },
        ]), (0, middlewares_1.asyncMiddleware)(this.sanitize));
        this.app.use((err, req, res, next) => {
            if (err.stack) {
                this.log.error('Request failed', 'url', req.url, 'stack', err.stack);
            }
            else {
                this.log.error('Request failed', 'url', req.url, 'error', err);
            }
            if (err.output) {
                return res.status(err.output.statusCode).json(err.output.payload);
            }
            return res.status(500).json(err);
        });
        this.createServer();
        const metrics = {
            durationHistogram: new promClient.Histogram({
                name: 'grafana_image_renderer_step_duration_seconds',
                help: 'duration histogram of browser steps for rendering an image labeled with: step',
                labelNames: ['step'],
                buckets: [0.1, 0.3, 0.5, 1, 3, 5, 10, 20, 30],
            }),
        };
        this.browser = (0, browser_1.createBrowser)(this.config.rendering, this.log, metrics);
        if (this.config.service.metrics.enabled) {
            const browserInfo = new promClient.Gauge({
                name: 'grafana_image_renderer_browser_info',
                help: "A metric with a constant '1 value labeled by version of the browser in use",
                labelNames: ['version'],
            });
            try {
                const browserVersion = await this.browser.getBrowserVersion();
                browserInfo.labels(browserVersion).set(1);
            }
            catch {
                this.log.error('Failed to get browser version');
                browserInfo.labels('unknown').set(0);
            }
        }
        await this.browser.start();
    }
    createServer() {
        const { protocol, host, port } = this.config.service;
        if (protocol === 'https') {
            const { certFile, certKey, minTLSVersion } = this.config.service;
            if (!certFile || !certKey) {
                throw new Error('No cert file or cert key provided, cannot start HTTPS server');
            }
            if (minTLSVersion && minTLSVersion !== 'TLSv1.2' && minTLSVersion !== 'TLSv1.3') {
                throw new Error('Only allowed TLS min versions are TLSv1.2 and TLSv1.3');
            }
            const options = {
                cert: fs.readFileSync(certFile),
                key: fs.readFileSync(certKey),
                maxVersion: 'TLSv1.3',
                minVersion: (minTLSVersion || 'TLSv1.2'),
            };
            this.server = https.createServer(options, this.app);
        }
        else {
            this.server = http.createServer(this.app);
        }
        if (host) {
            this.server.listen(port, host, () => {
                const info = this.server.address();
                this.log.info(`${protocol?.toUpperCase()} Server started, listening at ${protocol}://${host}:${info.port}`);
            });
        }
        else {
            this.server.listen(port, () => {
                const info = this.server.address();
                this.log.info(`${protocol?.toUpperCase()} Server started, listening at ${protocol}://localhost:${info.port}`);
            });
        }
    }
    close() {
        this.server.close();
    }
}
exports.HttpServer = HttpServer;
//# sourceMappingURL=http-server.js.map